package com.codigoparallevar.minicards.bridge;

import android.app.Notification;
import android.app.NotificationChannel;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Service;
import android.content.Context;
import android.content.Intent;
import android.os.AsyncTask;
import android.os.IBinder;
import android.os.Process;
import android.util.Log;
import android.widget.Toast;

import androidx.core.app.NotificationCompat;

import com.codigoparallevar.minicards.ConfigManager;
import com.codigoparallevar.minicards.R;
import com.codigoparallevar.minicards.types.functional.Tuple2;
import com.codigoparallevar.minicards.ui_helpers.DoAsync;
import com.programaker.api.ProgramakerApi;

public class ProgramakerBridgeService extends Service {
    public static final String BridgeUserNotificationChannel = "PROGRAMAKER_BRIDGE_USER_NOTIFICATION";
    public static final CharSequence BridgeUserNotificationChannelName = "User notifications";
    public static String BridgeUserVibrationNotificationChannel = "PROGRAMAKER_BRIDGE_USER_VIBRATION_NOTIFICATIONS";
    public static CharSequence BridgeUserVibrationNotificationChannelName = "User-triggered vibration";

    private static String BridgeStatusNotificationChannel = "PROGRAMAKER_BRIDGE_STATUS_NOTIFICATION";
    private static CharSequence BridgeStatusNotificationChannelName = "Programaker bridge status";

    private static final int BridgeStatusNotificationId = 9999;
    public static final String COMMAND_PROGRAMAKER_BRIDGE_STOP = "com.programaker.bridge.commands.stop";
    public static final String COMMAND_PROGRAMAKER_BRIDGE_START = "com.programaker.bridge.commands.start";

    private static final long WAIT_TIME_BEFORE_RESTART_MILLIS = 10000; // 10s
    private ProgramakerAndroidBridge bridge = null;
    private static final String LogTag = "PM BridgeService";
    private boolean stopped = false;

    @Override
    public void onCreate() {
        setBridgeStatusNotification(getString(R.string.bridge_service_not_started), true);
    }

    private void setBridgeStatusNotification(String title, boolean stopped) {
        NotificationManager notificationManager = (NotificationManager) getSystemService(Context.NOTIFICATION_SERVICE);
        assert notificationManager != null;

        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.O) {
            NotificationChannel channel = new NotificationChannel(
                    ProgramakerBridgeService.BridgeStatusNotificationChannel,
                    ProgramakerBridgeService.BridgeStatusNotificationChannelName,
                    NotificationManager.IMPORTANCE_DEFAULT);
            notificationManager.createNotificationChannel(channel);
        }

        Intent stopIntent = new Intent(this, ProgramakerBridgeService.class);
        stopIntent.setAction(COMMAND_PROGRAMAKER_BRIDGE_STOP);
        PendingIntent stopPendingIntent =
                PendingIntent.getService(this, 0, stopIntent, 0);

        Intent startIntent = new Intent(this, ProgramakerBridgeService.class);
        startIntent.setAction(COMMAND_PROGRAMAKER_BRIDGE_START);
        PendingIntent startPendingIntent =
                PendingIntent.getService(this, 0, startIntent, 0);

        NotificationCompat.Builder builder = new NotificationCompat
                .Builder(this, ProgramakerBridgeService.BridgeStatusNotificationChannel)
                .setContentTitle(title)
                .setPriority(NotificationCompat.PRIORITY_DEFAULT);


        if (stopped) {
            builder.addAction(R.drawable.ic_start, getString(R.string.start_bridge), startPendingIntent)
                    .setSmallIcon(R.drawable.ic_vector_stopped_icon);
        }
        else {
            builder.addAction(R.drawable.ic_cancel, getString(R.string.stop_bridge), stopPendingIntent)
                    .setSmallIcon(R.drawable.ic_vector_icon);
        }

        Notification notification = builder.build();

        if (stopped) {
            notificationManager.notify(BridgeStatusNotificationId, notification);
        }
        else {
            this.startForeground(BridgeStatusNotificationId, notification);
        }
    }

    @Override
    public int onStartCommand(Intent intent, int flags, int startId) {
        String action = intent.getAction();
        if (action == null) {
            action = "";
        }

        if (action.equals(COMMAND_PROGRAMAKER_BRIDGE_STOP)) {
            this.stopSelf();
        }
        else {
            connectBridge();
        }

        // If we get killed, after returning from here, restart
        return START_STICKY;
    }

    private void connectBridge() {
        stopped = false;
        ConfigManager config = new ConfigManager(this);

        String token = config.getToken();
        if (token == null) {
            Toast.makeText(this, "Cannot start bridge", Toast.LENGTH_SHORT).show();
            Log.e(LogTag, "Cannot start bridge: Token is null");
            this.stopSelf();
        }

        if (ProgramakerBridgeService.this.bridge != null) {
            // Toast.makeText(this, "Bridge already started", Toast.LENGTH_SHORT).show();
            Log.w(LogTag, "Bridge already started (not null)");
        }

        // Start up the thread running the service. Note that we create a
        // separate thread because the service normally runs in the process's
        // main thread, which we don't want to block. We also make it
        // background priority so CPU-intensive work doesn't disrupt our UI.
        Thread thread = new Thread(() -> {
            try {
                ProgramakerApi api = new ProgramakerApi();
                api.setToken(token);
                String userId = api.getUserId();

                String bridgeIdCheck = config.getBridgeId();
                if (bridgeIdCheck == null) {
                    bridgeIdCheck = api.createBridge(ProgramakerAndroidBridge.GetBridgeName(this));
                    config.setBridgeId(bridgeIdCheck);
                }
                final String bridgeId = bridgeIdCheck;

                ProgramakerBridgeService.this.bridge = ProgramakerAndroidBridge.configure(
                        this,
                        userId,
                        bridgeId);
                ProgramakerBridgeService.this.bridge.start(
                        () -> { // On ready
                            setBridgeStatusNotification(getString(R.string.bridge_service_online), false);
                            if (config.getBridgeConnectionId() == null) {
                                new DoAsync().executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR,
                                        new Tuple2<>(
                                                () -> {
                                                    boolean established = api.establishConnection(bridgeId);
                                                    if (!established) {
                                                        Log.e(LogTag, "Error establishing connection to bridge");
                                                    }
                                                },
                                                ex -> {
                                                    Log.e(LogTag, "Error establishing bridge connection: " + ex, ex);
                                                    ProgramakerBridgeService.this.stopSelf();
                                                }
                                        )
                                );
                            }
                        },
                        () -> { // On completed
                            ProgramakerBridgeService.this.bridge = null;
                            onBridgeFailedAfterConnected();
                        });
            }
            catch (Throwable ex) {
                Log.e(LogTag, "Error on bridge", ex);
                ProgramakerBridgeService.this.bridge = null;
            }
        }, "ServiceStartArguments");
        thread.setPriority(Process.THREAD_PRIORITY_BACKGROUND);

        setBridgeStatusNotification(getString(R.string.bridge_service_starting), false);
        thread.start();
    }

    private void onBridgeFailedAfterConnected() {
        if (!stopped) {
            Log.e(LogTag, "Bridge stopped after connected. Waiting 10s then restarting");
            setBridgeStatusNotification(getString(R.string.bridge_service_failed_restarting), false);
            try {
                Thread.sleep(WAIT_TIME_BEFORE_RESTART_MILLIS);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            connectBridge();
        }
    }

    @Override
    public IBinder onBind(Intent intent) {
        // We don't provide binding, so return null
        return null;
    }

    @Override
    public void onDestroy() {
        stopped = true;
        ProgramakerAndroidBridge bridge = this.bridge;
        if (bridge != null) {
            bridge.stop();
        }
        setBridgeStatusNotification(getString(R.string.bridge_service_failed_stopping), true);
    }
}
