package com.programaker.api

import android.os.Build
import android.util.Log
import com.google.gson.Gson
import com.google.gson.GsonBuilder
import com.programaker.api.data.ProgramakerBridgeInfo
import com.programaker.api.data.ProgramakerCustomBlock
import com.programaker.api.data.ProgramakerFunctionCallResult
import com.programaker.api.data.api_results.*
import com.programaker.api.exceptions.ProgramakerLoginRequiredException
import com.programaker.api.exceptions.ProgramakerProtocolException
import com.programaker.api.exceptions.TokenNotFoundException
import org.json.JSONArray
import org.json.JSONObject
import java.io.DataOutputStream
import java.io.FileNotFoundException
import java.io.InputStream
import java.io.InputStreamReader
import java.lang.reflect.Type
import java.net.HttpURLConnection
import java.net.URL
import java.net.URLConnection


class ProgramakerApi(private val ApiRoot: String="https://programaker.com/api") {
    private val LogTag: String = "ProgramakerApi"

    private var userId: String? = null
    private var userName: String? = null
    var token: String? = null

    // API
    fun check(): Boolean {
        val conn = URL(getCheckUrl()).openConnection() as HttpURLConnection
        if (conn == null){
            Log.e(LogTag, "URL Connection not established, set to NULL")
            return false
        }
        try {
            addAuthHeader(conn)
        } catch(ex: TokenNotFoundException) {
            return false
        }

        val result: ProgramakerCheckResult
        try {
            result = parseJson(conn.inputStream, ProgramakerCheckResult::class.java)
        } catch(ex: JsonParseException) {
            ex.logError(LogTag)
            return false
        } catch (ex: FileNotFoundException) {
            ex.logError(LogTag)
            return false
        }

        if (result.success) {
            this.userId = result.user_id;
            this.userName = result.username;
        }

        return result.success
    }


    fun login(user: String, password: String): String {
        val conn = URL(getLoginUrl()).openConnection() as HttpURLConnection
        conn.setRequestProperty("Content-Type", "application/json")

        conn.requestMethod = "POST";
        conn.doOutput = true;

        val postData = JSONObject()
        postData.put("username", user)
        postData.put("password", password)

        val wr = DataOutputStream(conn.outputStream)
        wr.writeBytes(postData.toString());
        wr.flush();
        wr.close();

        val result: ProgramakerLoginResult
        result = parseJson(conn.inputStream, ProgramakerLoginResult::class.java)
        return result.token
    }

    fun fetchCustomBlocks(): List<ProgramakerBridgeCustomBlockResult> {
        val conn = URL(getCustomBlocksUrl()).openConnection() as HttpURLConnection

        addAuthHeader(conn)

        val result: ProgramakerGetCustomBlocksResult

        try {
            val builder = GsonBuilder()
            builder.registerTypeAdapter(ProgramakerGetCustomBlocksResult::class.java, ProgramakerGetCustomBlocksResultTypeAdapter())

            val gson = builder.create()
            val reader = InputStreamReader(conn.inputStream)

            result = gson.fromJson(reader, ProgramakerGetCustomBlocksResult::class.java)
        } catch(ex: JsonParseException) {
            ex.logError(LogTag)
            throw ProgramakerProtocolException()
        } catch (ex: FileNotFoundException) {
            ex.logError(LogTag)
            throw ProgramakerProtocolException()
        }
        catch (ex: Exception) {
            Log.e(LogTag, "Unexpected exception: " + ex, ex)
            throw ProgramakerProtocolException()
        }
        return result.result
    }

    fun fetchConnectedBridges(): List<ProgramakerBridgeInfo> {
        val conn = URL(getConnectedBridgesUrl()).openConnection() as HttpURLConnection

        addAuthHeader(conn)

        val result: ProgramakerGetBridgeInfoResult

        try {
            val builder = GsonBuilder()
            builder.registerTypeAdapter(ProgramakerGetBridgeInfoResult::class.java, ProgramakerGetBridgeInfoTypeAdapter())

            val gson = builder.create()
            val reader = InputStreamReader(conn.inputStream)

            result = gson.fromJson(reader, ProgramakerGetBridgeInfoResult::class.java)
        } catch(ex: JsonParseException) {
            ex.logError(LogTag)
            throw ProgramakerProtocolException()
        } catch (ex: FileNotFoundException) {
            ex.logError(LogTag)
            throw ProgramakerProtocolException()
        }
        return result.result
    }

    fun callBlock(block: ProgramakerCustomBlock, arguments: List<String>): ProgramakerFunctionCallResult {
        val conn = URL(getBlockUrl(block)).openConnection() as HttpURLConnection
        conn.setRequestProperty("Content-Type", "application/json")
        addAuthHeader(conn)

        conn.requestMethod = "POST";
        conn.doOutput = true;

        val postData = JSONObject(hashMapOf(
                "arguments" to JSONArray(arguments)
        ) as Map<*, *>)

        val wr = DataOutputStream(conn.outputStream)
        wr.writeBytes(postData.toString());
        wr.flush();
        wr.close();

        val result: ProgramakerFunctionCallResult
        result = parseJson(conn.inputStream, ProgramakerFunctionCallResult::class.java)
        return result
    }

    fun createBridge(name: String): String {
        val conn = URL(getCreateBridgeUrl()).openConnection() as HttpURLConnection
        conn.setRequestProperty("Content-Type", "application/json")
        addAuthHeader(conn)

        conn.requestMethod = "POST";
        conn.doOutput = true;

        val postData = JSONObject(hashMapOf(
                "name" to name
        ) as Map<*, *>)

        val wr = DataOutputStream(conn.outputStream)
        wr.writeBytes(postData.toString());
        wr.flush();
        wr.close();

        val result: ProgramakerCreateBridgeResult
        result = parseJson(conn.inputStream, ProgramakerCreateBridgeResult::class.java)
        return result.getBridgeId()
    }

    fun establishConnection(bridgeId: String): Boolean {
        // NOTE: This establishes a connection to a bridge as long as it can be done without interaction

        // Prepare connection
        val prepareConn = URL(getPrepareConnectionUrl(bridgeId)).openConnection() as HttpURLConnection
        addAuthHeader(prepareConn)

        val prepareResult: ProgramakerPrepareConnectionResult
        prepareResult = parseJson(prepareConn.inputStream, ProgramakerPrepareConnectionResult::class.java)
        if (prepareResult.type != "direct") {
            throw Exception("Expected 'direct' connection type, found '${prepareResult.type}'")
        }

        // Establish connection
        val establishConn = URL(getEstablishConnectionUrl(bridgeId)).openConnection() as HttpURLConnection
        establishConn.setRequestProperty("Content-Type", "application/json")
        addAuthHeader(establishConn)

        establishConn.requestMethod = "POST";
        establishConn.doOutput = true;

        val establishWrite = DataOutputStream(establishConn.outputStream)
        establishWrite.writeBytes(JSONObject().toString());
        establishWrite.flush();
        establishWrite.close();

        val establishResult: ProgramakerEstablishDirectConnectionResult
        establishResult = parseJson(establishConn.inputStream, ProgramakerEstablishDirectConnectionResult::class.java)
        return establishResult.success
    }

    // Initialization
    init {
        // Disable connection reuse if necessary
        // HTTP connection reuse which was buggy pre-froyo
        if (Build.VERSION.SDK.toInt() < Build.VERSION_CODES.FROYO) {
            System.setProperty("http.keepAlive", "false")
        }
    }

    // Private functions
    private fun getCheckUrl(): String {
        return "$ApiRoot/v0/sessions/check"
    }

    private fun getLoginUrl(): String {
        return "$ApiRoot/v0/sessions/login"
    }

    private fun getCustomBlocksUrl(): String {
        this.withUserName()
        return "$ApiRoot/v0/users/$userName/custom-blocks/"
    }

    private fun getConnectedBridgesUrl(): String {
        this.withUserName()
        return "$ApiRoot/v0/users/$userName/bridges/"
    }

    private fun getBlockUrl(block: ProgramakerCustomBlock): String {
        this.withUserId()
        return "$ApiRoot/v0/users/id/$userId/bridges/id/${block.bridge_id}/functions/${block.function_name}"
    }

    private fun getCreateBridgeUrl(): String {
        this.withUserName()
        return "$ApiRoot/v0/users/$userName/bridges"
    }

    private fun getPrepareConnectionUrl(bridgeId: String): String {
        this.withUserName()
        return "$ApiRoot/v0/users/$userName/services/id/$bridgeId/how-to-enable"
    }

    private fun getEstablishConnectionUrl(bridgeId: String): String {
        this.withUserName()
        return "$ApiRoot/v0/users/$userName/services/id/$bridgeId/register"
    }

    fun getUserId(): String? {
        this.withUserId()
        return userId;
    }

    private fun withUserName() {
        if (userName == null) {
            if (token == null) {
                throw ProgramakerLoginRequiredException()
            }

            this.check();
            if (userName == null) {
                throw ProgramakerProtocolException()
            }
        }
    }

    private fun withUserId() {
        if (userId == null) {
            if (token == null) {
                throw ProgramakerLoginRequiredException()
            }

            this.check();
            if (userId == null) {
                throw ProgramakerProtocolException()
            }
        }
    }

    private fun addAuthHeader(conn: URLConnection) {
        if (token != null) {
            conn.setRequestProperty("Authorization", token)
        }
        else {
            throw TokenNotFoundException()
        }
    }

    private fun <T>parseJson(content: InputStream, resultClass: Type): T {
        val reader = InputStreamReader(content)
        val gson = Gson()
        return gson.fromJson(reader, resultClass)
    }
}

private fun FileNotFoundException.logError(tag: String) {
    Log.e(tag, "Cannot open: ${this.message}")
}

class JsonParseException(private val content: String, private val cls: Type) : Exception() {
    fun logError(tag: String) {
        Log.e(tag, "Cannot JSON parse: ${this.content} as ${this.cls}", this)
    }
}
